@@ -1,12 +1,16 @@
-from ..region import Section
+from __future__ import annotations
+
+from cle.backends.region import Section
+
 
 class PESection(Section):
     """
     Represents a section for the PE format.
     """
+
     def __init__(self, pe_section, remap_offset=0):
         super().__init__(
-            pe_section.Name.decode(),
+            pe_section.Name.decode("latin-1"),  # ensure all bytes can be decoded
             pe_section.PointerToRawData,
             pe_section.VirtualAddress + remap_offset,
             pe_section.Misc_VirtualSize,
