@@ -1,23 +1,19 @@
+from __future__ import annotations
+
+import itertools
 import logging
 import os
 import platform
 import sys
 from collections import OrderedDict
+from collections.abc import Iterable, Iterator
 from pathlib import Path
 from typing import (
     TYPE_CHECKING,
     Any,
     BinaryIO,
-    Dict,
-    Iterable,
-    Iterator,
-    List,
     Literal,
-    Optional,
-    Set,
-    Type,
     TypeVar,
-    Union,
 )
 
 import archinfo
@@ -59,14 +55,14 @@ class Loader:
 
     def __init__(
         self,
-        main_binary: Union[str, BinaryIO, Path, Backend],
+        main_binary: str | BinaryIO | Path | Backend,
         auto_load_libs: bool = True,
         concrete_target=None,
-        force_load_libs: Iterable[Union[str, BinaryIO, Path]] = (),
+        force_load_libs: Iterable[str | BinaryIO | Path] = (),
         skip_libs: Iterable[str] = (),
-        main_opts: Optional[Dict[str, Any]] = None,
-        lib_opts: Optional[Dict[str, Dict[str, Any]]] = None,
-        ld_path: Iterable[Union[str, Path]] = (),
+        main_opts: dict[str, Any] | None = None,
+        lib_opts: dict[str, dict[str, Any]] | None = None,
+        ld_path: Iterable[str | Path] = (),
         use_system_libs: bool = True,
         ignore_import_version_numbers: bool = True,
         case_insensitive: bool = False,
@@ -76,8 +72,8 @@ class Loader:
         perform_relocations: bool = True,
         load_debug_info: bool = False,
         page_size: int = 0x1,
-        preload_libs: Iterable[Union[str, BinaryIO, Path]] = (),
-        arch: Union[archinfo.Arch, str, None] = None,
+        preload_libs: Iterable[str | BinaryIO | Path] = (),
+        arch: archinfo.Arch | str | None = None,
     ):
         """
         :param main_binary:         The path to the main binary you're loading, or a file-like object with the binary
@@ -147,7 +143,7 @@ class Loader:
 
         self._auto_load_libs = auto_load_libs
         self._load_debug_info = load_debug_info
-        self._satisfied_deps: Dict[str, Union[Literal[False], Backend]] = {x: False for x in skip_libs}
+        self._satisfied_deps: dict[str, Literal[False] | Backend] = {x: False for x in skip_libs}
         self._main_opts = {} if main_opts is None else main_opts
         self._lib_opts = {} if lib_opts is None else lib_opts
         self._custom_ld_path = [ld_path] if isinstance(ld_path, str) else ld_path
@@ -175,12 +171,13 @@ class Loader:
         self.aslr = aslr
         self.page_size = page_size
         self._memory = None
+        self._original_main_object = None
         self._main_object = None
         self._tls = None
-        self._kernel_object: Optional[KernelObject] = None
-        self._extern_object: Optional[ExternObject] = None
+        self._kernel_object: KernelObject | None = None
+        self._extern_object: ExternObject | None = None
         self.shared_objects = OrderedDict()
-        self.all_objects: List[Backend] = []
+        self.all_objects: list[Backend] = []
         self.requested_names = set()
         if arch is not None:
             self._main_opts.update({"arch": arch})
@@ -207,6 +204,13 @@ class Loader:
             raise ValueError("Cannot access main_object before loading is complete")
         return result
 
+    @property
+    def original_main_object(self) -> Backend:
+        result = self._original_main_object
+        if result is None:
+            raise ValueError("Cannot access original_main_object before loading is complete")
+        return result
+
     @property
     def memory(self) -> Clemory:
         result = self._memory
@@ -215,7 +219,7 @@ class Loader:
         return result
 
     @property
-    def tls(self) -> "ThreadManager":
+    def tls(self) -> ThreadManager:
         result = self._tls
         if result is None:
             raise ValueError("Cannot access tls before loading is complete")
@@ -245,7 +249,7 @@ class Loader:
         return self.all_objects[0].min_addr
 
     @property
-    def initializers(self) -> List[int]:
+    def initializers(self) -> list[int]:
         """
         Return a list of all the initializers that should be run before execution reaches the entry point, in the order
         they should be run.
@@ -253,7 +257,7 @@ class Loader:
         return sum((x.initializers for x in self.all_objects), [])
 
     @property
-    def finalizers(self) -> List[int]:
+    def finalizers(self) -> list[int]:
         """
         Return a list of all the finalizers that should be run before the program exits.
         I'm not sure what order they should be run in.
@@ -261,7 +265,7 @@ class Loader:
         return sum((x.finalizers for x in self.all_objects), [])
 
     @property
-    def linux_loader_object(self) -> Optional[Backend]:
+    def linux_loader_object(self) -> Backend | None:
         """
         If the linux dynamic loader is present in memory, return it
         """
@@ -273,7 +277,7 @@ class Loader:
         return None
 
     @property
-    def elfcore_object(self) -> Optional[ELFCore]:
+    def elfcore_object(self) -> ELFCore | None:
         """
         If a corefile was loaded, this returns the actual core object instead of the main binary
         """
@@ -310,7 +314,7 @@ class Loader:
         return self._extern_object
 
     @property
-    def kernel_object(self) -> "KernelObject":
+    def kernel_object(self) -> KernelObject:
         """
         Return the object used to provide addresses to syscalls.
 
@@ -322,21 +326,21 @@ class Loader:
         return self._kernel_object
 
     @property
-    def all_elf_objects(self) -> List[MetaELF]:
+    def all_elf_objects(self) -> list[MetaELF]:
         """
         Return a list of every object that was loaded from an ELF file.
         """
         return [o for o in self.all_objects if isinstance(o, MetaELF)]
 
     @property
-    def all_pe_objects(self) -> List[PE]:
+    def all_pe_objects(self) -> list[PE]:
         """
         Return a list of every object that was loaded from an ELF file.
         """
         return [o for o in self.all_objects if isinstance(o, PE)]
 
     @property
-    def missing_dependencies(self) -> Set[str]:
+    def missing_dependencies(self) -> set[str]:
         """
         Return a set of every name that was requested as a shared object dependency but could not be loaded
         """
@@ -365,7 +369,7 @@ class Loader:
             if not sym.name or sym.is_import:
                 idx -= 1
                 continue
-            options.append((sym.relative_addr, "%s+" % sym.name))
+            options.append((sym.relative_addr, f"{sym.name}+"))
             break
 
         if isinstance(o, ELF):
@@ -374,7 +378,7 @@ class Loader:
             except ValueError:
                 pass
             else:
-                options.append((plt_addr, "PLT.%s+" % plt_name))
+                options.append((plt_addr, f"PLT.{plt_name}+"))
 
         options.append((0, "offset "))
 
@@ -392,7 +396,7 @@ class Loader:
 
     # Search functions
 
-    def find_object(self, spec: Union[Backend, str], extra_objects: Iterable[Backend] = ()) -> Optional[Backend]:
+    def find_object(self, spec: Backend | str, extra_objects: Iterable[Backend] = ()) -> Backend | None:
         """
         If the given library specification has been loaded, return its object, otherwise return None.
         """
@@ -419,7 +423,7 @@ class Loader:
 
         return None
 
-    def find_object_containing(self, addr: int, membership_check: bool = True) -> Optional[Backend]:
+    def find_object_containing(self, addr: int, membership_check: bool = True) -> Backend | None:
         """
         Return the object that contains the given address, or None if the address is unmapped.
 
@@ -440,7 +444,7 @@ class Loader:
                 self._last_object = obj_
                 return obj_
             else:
-                raise CLEError("Unsupported memory type %s" % type(obj_.memory))
+                raise CLEError(f"Unsupported memory type {type(obj_.memory)}")
 
         # check the cache first
         if self._last_object is not None and self._last_object.min_addr <= addr <= self._last_object.max_addr:
@@ -468,7 +472,7 @@ class Loader:
             return obj
         return _check_object_memory(obj)
 
-    def find_segment_containing(self, addr: int, skip_pseudo_objects: bool = True) -> Optional["Segment"]:
+    def find_segment_containing(self, addr: int, skip_pseudo_objects: bool = True) -> Segment | None:
         """
         Find the section object that the address belongs to.
 
@@ -490,7 +494,7 @@ class Loader:
 
         return obj.find_segment_containing(addr)
 
-    def find_section_containing(self, addr: int, skip_pseudo_objects=True) -> Optional["Section"]:
+    def find_section_containing(self, addr: int, skip_pseudo_objects=True) -> Section | None:
         """
         Find the section object that the address belongs to.
 
@@ -512,7 +516,7 @@ class Loader:
 
         return obj.find_section_containing(addr)
 
-    def find_loadable_containing(self, addr: int, skip_pseudo_objects=True) -> Optional["Region"]:
+    def find_loadable_containing(self, addr: int, skip_pseudo_objects=True) -> Region | None:
         """
         Find the section or segment object the address belongs to. Sections will only be used if the corresponding
         object does not have segments.
@@ -533,7 +537,7 @@ class Loader:
 
         return obj.find_loadable_containing(addr)
 
-    def find_section_next_to(self, addr: int, skip_pseudo_objects=True) -> Optional["Section"]:
+    def find_section_next_to(self, addr: int, skip_pseudo_objects=True) -> Section | None:
         """
         Find the next section after the given address.
 
@@ -555,7 +559,7 @@ class Loader:
 
         return obj.sections.find_region_next_to(addr)
 
-    def find_symbol(self, thing, fuzzy=False) -> Optional[Symbol]:
+    def find_symbol(self, thing, fuzzy=False) -> Symbol | None:
         """
         Search for the symbol with the given name or address.
 
@@ -657,7 +661,7 @@ class Loader:
 
                 yield sym
 
-    def find_plt_stub_name(self, addr: int) -> Optional[str]:
+    def find_plt_stub_name(self, addr: int) -> str | None:
         """
         Return the name of the PLT stub starting at ``addr``.
         """
@@ -666,7 +670,7 @@ class Loader:
             return so.reverse_plt.get(addr, None)
         return None
 
-    def find_relevant_relocations(self, name: str) -> Iterator["Relocation"]:
+    def find_relevant_relocations(self, name: str) -> Iterator[Relocation]:
         """
         Iterate through all the relocations referring to the symbol with the given ``name``
         """
@@ -786,8 +790,15 @@ class Loader:
 
             if self._main_object is None:
                 # this is technically the first place we can start to initialize things based on platform
-                self._main_object = obj
-                self._memory = Clemory(obj.arch, root=True)
+                if obj.force_main_object is not None:
+                    self._original_main_object = obj
+                    self._main_object = obj.force_main_object
+                    self._main_object.is_main_bin = True
+                    # just to be safe, we clear obj.force_main_object here
+                    obj.force_main_object = None
+                else:
+                    self._original_main_object = self._main_object = obj
+                self._memory = Clemory(self._main_object.arch, root=True)
 
                 chk_obj = (
                     self.main_object
@@ -795,15 +806,15 @@ class Loader:
                     else self.main_object.child_objects[0]
                 )
                 if isinstance(chk_obj, ELFCore):
-                    self._tls = ELFCoreThreadManager(self, obj.arch)
-                elif isinstance(obj, Minidump):
-                    self._tls = MinidumpThreadManager(self, obj.arch)
+                    self._tls = ELFCoreThreadManager(self, self._main_object.arch)
+                elif isinstance(self._main_object, Minidump):
+                    self._tls = MinidumpThreadManager(self, self._main_object.arch)
                 elif isinstance(chk_obj, MetaELF):
-                    self._tls = ELFThreadManager(self, obj.arch)
+                    self._tls = ELFThreadManager(self, self._main_object.arch)
                 elif isinstance(chk_obj, PE):
-                    self._tls = PEThreadManager(self, obj.arch)
+                    self._tls = PEThreadManager(self, self._main_object.arch)
                 else:
-                    self._tls = ThreadManager(self, obj.arch)
+                    self._tls = ThreadManager(self, self._main_object.arch)
 
             elif is_preloading:
                 self.preload_libs.append(obj)
@@ -931,7 +942,7 @@ class Loader:
 
         return objects
 
-    def _load_object_isolated(self, spec):
+    def _load_object_isolated(self, spec, obj_ident: str | None = None):
         """
         Given a partial specification of a dependency, this will return the loaded object as a backend instance.
         It will not touch any loader-global data.
@@ -953,16 +964,18 @@ class Loader:
             binary_stream = spec.open("rb")
             close = True
         else:
-            raise CLEError("Bad library specification: %s" % spec)
+            raise CLEError(f"Bad library specification: {spec}")
 
         try:
             # STEP 2: collect options
             if self._main_object is None:
                 options = dict(self._main_opts)
             else:
-                for ident in self._possible_idents(
-                    binary_stream if binary is None else binary
+                for ident in itertools.chain(
+                    [obj_ident], self._possible_idents(binary_stream if binary is None else binary)
                 ):  # also allowed to cheat
+                    if ident is None:
+                        continue
                     if ident in self._lib_opts:
                         options = dict(self._lib_opts[ident])
                         break
@@ -976,7 +989,7 @@ class Loader:
                 backend_cls = self._static_backend(binary_stream if binary is None else binary)
             if backend_cls is None:
                 raise CLECompatibilityError(
-                    "Unable to find a loader backend for %s.  Perhaps try the 'blob' loader?" % spec
+                    f"Unable to find a loader backend for {spec}.  Perhaps try the 'blob' loader?"
                 )
 
             # STEP 4: LOAD!
@@ -989,7 +1002,7 @@ class Loader:
             if close:
                 binary_stream.close()
 
-    def _map_object(self, obj: "Backend"):
+    def _map_object(self, obj: Backend):
         """
         This will integrate the object into the global address space, but will not perform relocations.
         """
@@ -1101,7 +1114,7 @@ class Loader:
 
             return path
 
-        raise CLEFileNotFoundError("Could not find file %s" % spec)
+        raise CLEFileNotFoundError(f"Could not find file {spec}")
 
     def _possible_paths(self, spec):
         """
@@ -1261,7 +1274,7 @@ class Loader:
             for name in self._possible_idents(spec, lowercase=True):
                 yield name.lower()
 
-    def _static_backend(self, spec, ignore_hints=False) -> Optional[Type[Backend]]:
+    def _static_backend(self, spec, ignore_hints=False) -> type[Backend] | None:
         """
         Returns the correct loader for the file at `spec`.
         Returns None if it's a blob or some unknown type.
@@ -1276,16 +1289,14 @@ class Loader:
                     pass
 
         with stream_or_path(spec) as stream:
-            for rear in ALL_BACKENDS.values():
+            for rear in [bk for bk in ALL_BACKENDS.values() if bk is not Blob] + [Blob]:
                 if rear.is_default and rear.is_compatible(stream):
                     return rear
 
         return None
 
     @staticmethod
-    def _backend_resolver(
-        backend: Union[str, Type[Backend]], default: Optional[T] = None
-    ) -> Union[Type[Backend], Optional[T]]:
+    def _backend_resolver(backend: str | type[Backend], default: T | None = None) -> type[Backend] | T | None:
         if isinstance(backend, type) and issubclass(backend, Backend):
             return backend
         elif backend in ALL_BACKENDS:
@@ -1293,13 +1304,13 @@ class Loader:
         elif backend is None:
             return default
         else:
-            raise CLEError("Invalid backend: %s" % backend)
+            raise CLEError(f"Invalid backend: {backend}")
 
     #
     # Memory data loading methods
     #
 
-    def fast_memory_load_pointer(self, addr: int, size: Optional[int] = None) -> Optional[int]:
+    def fast_memory_load_pointer(self, addr: int, size: int | None = None) -> int | None:
         """
         Perform a fast memory loading of a pointer.
 
